
/*
  Kjetil Matheussen, 2005/2006.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sndfile.h>
#include <pthread.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>



#define JC_MAX(a,b) (((a)>(b))?(a):(b))
#define JC_MIN(a,b) (((a)<(b))?(a):(b))

#define OPTARGS_BEGIN(das_usage) {int lokke;const char *usage=das_usage;for(lokke=1;lokke<argc;lokke++){char *a=argv[lokke];if(!strcmp("--help",a)||!strcmp("-h",a)){printf(usage);return 0
#define OPTARG(name,name2) }else if(!strcmp(name,a)||!strcmp(name2,a)){
#define OPTARG_GETINT() atoi(argv[++lokke])
#define OPTARG_GETFLOAT() atof(argv[++lokke])
#define OPTARG_GETSTRING() argv[++lokke]
#define OPTARG_LAST() }else if(lokke==argc-1){lokke--;
#define OPTARGS_END }else{fprintf(stderr,usage);return(-1);}}}



/* Arguments and their default values */
static jack_nframes_t rb_size=16384*32*8*sizeof(jack_default_audio_sample_t);
static jack_client_t *client=NULL;
static unsigned int channels=2;
static int bitdepth=0;
static char *filename=NULL;
static int leading_zeros=0;


/* JACK data */
static jack_port_t **ports;
typedef jack_default_audio_sample_t sample_t;
static unsigned long long overruns = 0;
static unsigned long long overruns_serious = 0;
static int jack_buffer_size;

/* Disk thread */
static pthread_t disk_thread={0};
static pthread_mutex_t disk_thread_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  data_ready = PTHREAD_COND_INITIALIZER;

/* Synchronization between process thread and disk thread. */
static volatile int can_capture=0;
static volatile int can_process=0;
static volatile int is_running=1;
static jack_ringbuffer_t *rb;

/* Jack connecting thread. */
static pthread_t connect_thread={0} ;
static pthread_mutex_t connect_thread_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  connect_ready = PTHREAD_COND_INITIALIZER;





/////////////////////////////////////////////////////////////////////
//////////////////////// BUFFERS ////////////////////////////////////
/////////////////////////////////////////////////////////////////////


static sample_t **buffers;
static sample_t *empty_buffer;

static int next_free_buffer=0;
static int num_free_buffers;
static int num_buffers;
static int buffer_size;
static int buffer_size_in_bytes;

void buffers_init(){
  static sample_t *das_buffer;
  int lokke;
  buffer_size=channels*jack_buffer_size;
  buffer_size_in_bytes=buffer_size*sizeof(sample_t);
  num_buffers=3+(rb_size/buffer_size_in_bytes);
  num_free_buffers=num_buffers-2;

  das_buffer=calloc(num_buffers,buffer_size_in_bytes);
  buffers=calloc(sizeof(sample_t),num_buffers);

  for(lokke=0;lokke<num_buffers;lokke++){
    buffers[lokke]=das_buffer+(lokke*buffer_size);
  }
  //printf("Buffers[lokke](size) %d %d %d\n",buffers[1]-buffers[0]);

  empty_buffer=calloc(1,buffer_size_in_bytes);
}

sample_t *buffer_get(void){
  if(num_free_buffers==0){
    return empty_buffer;
  }else{
    sample_t *ret=buffers[next_free_buffer];
    num_free_buffers--;
    next_free_buffer++;
    if(next_free_buffer==num_buffers)
      next_free_buffer=0;
    return ret;
  }
}

void buffer_put(sample_t *buffer){
  if(buffer!=empty_buffer)
    num_free_buffers++;
}




/////////////////////////////////////////////////////////////////////
//////////////////////// DISK ///////////////////////////////////////
/////////////////////////////////////////////////////////////////////

void *disk_thread_func (void *arg){
        SNDFILE *soundfile;
	SF_INFO sf_info={0};
	int short_mask;

	unsigned long long lastchecked_overruns=0;
	unsigned long long lastchecked_overruns_serious=0;

	{
	  sf_info.samplerate = jack_get_sample_rate (client);
	  sf_info.channels = channels;
	  
	  switch (bitdepth) {
	  case 8: short_mask = SF_FORMAT_PCM_U8;
	    break;
	  case 16: short_mask = SF_FORMAT_PCM_16;
	    break;
	  case 24: short_mask = SF_FORMAT_PCM_24;
	    break;
	  case 32: short_mask = SF_FORMAT_PCM_32;
	    break;
	  default: short_mask = SF_FORMAT_FLOAT;
	    break;
	  }		 
	  sf_info.format = SF_FORMAT_WAV|short_mask;
	  
	  if ((soundfile = sf_open (filename, SFM_WRITE, &sf_info)) == NULL) {
	    char errstr[256];
	    sf_error_str (0, errstr, sizeof (errstr) - 1);
	    fprintf (stderr, "cannot open sndfile \"%s\" for output (%s)\n", filename, errstr);
	    jack_client_close (client);
	    exit (1);
	  }
	}


	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	pthread_mutex_lock (&disk_thread_lock);

	for(;;){
	  if (overruns > lastchecked_overruns) {
	    fprintf (stderr,
		     "jack_capture failed with %ld overruns.\n", (long)(overruns-lastchecked_overruns));
	    fprintf (stderr, " try a bigger buffer than -B %"
		     PRIu32 ".\n", rb_size);
	    fprintf(stderr,"Continue recording...\n");
	    lastchecked_overruns=overruns;
	  }
	  if (overruns_serious > lastchecked_overruns_serious) {
	    fprintf (stderr,
		     "jack_capture failed seriously regarding buffers. Your soundfile will so far be %llu frames shorter than the recording.\n", 
		     overruns_serious*jack_buffer_size);
	    fprintf (stderr, " try a bigger buffer than -B %"
		     PRIu32 ".\n", rb_size);
	    fprintf(stderr,"Continue recording...\n");
	    lastchecked_overruns_serious=overruns_serious;
	  }

	  while (can_capture && jack_ringbuffer_read_space (rb) >= sizeof(sample_t*)) {
	    sample_t *buffer=NULL;
	    jack_ringbuffer_read(rb,(void*)&buffer,sizeof(sample_t*));
	    if(sf_writef_float(soundfile,buffer,buffer_size/channels) != buffer_size/channels) {
	      char errstr[256];
	      sf_error_str (0, errstr, sizeof (errstr) - 1);
	      fprintf (stderr,
		       "cannot write sndfile (%s)\n",
		       errstr);
	      goto done;
	    }
	    buffer_put(buffer);
	  }
	  
	  if (!is_running) {
	    goto done;
	  }

	  /* wait until process() signals more data */
	  pthread_cond_wait (&data_ready, &disk_thread_lock);
	}


 done:

	{
	  sf_close (soundfile);
	  if (overruns > 0) {
	    fprintf (stderr,
		     "jack_capture failed with a total of %llu overruns.\n", overruns);
	    fprintf (stderr, " try a bigger buffer than -B %"
		     PRIu32 ".\n", rb_size);
	  }
	}

	pthread_mutex_unlock (&disk_thread_lock);
	printf ("disk thread finished\n");
	return 0;
}

void setup_disk_thread (void){
  pthread_create (&disk_thread, NULL, disk_thread_func, NULL);
}

void start_disk_thread (void){
  can_capture = 1;
}

void stop_disk_thread(void){
 /* Wake up disk thread. (no trylock) (isrunning==0)*/
  pthread_cond_signal (&data_ready);
  pthread_join (disk_thread, NULL);
}




/////////////////////////////////////////////////////////////////////
//////////////////////// JACK ///////////////////////////////////////
/////////////////////////////////////////////////////////////////////	
int process (jack_nframes_t nframes, void *arg){
	int chn;
	jack_default_audio_sample_t *in[channels];


	/* Do nothing until we're ready to begin. */
	if ((!can_process) || (!can_capture))
	  return 0;


	// get jack buffers
	for (chn = 0; chn < channels; chn++)
	  in[chn] = jack_port_get_buffer (ports[chn], nframes);


	// jack buffers -> disk buffer
	{
	  sample_t *buffer=buffer_get();
	  int i,chn,pos=0;
	  if(buffer==empty_buffer){
	    overruns++;
	  }else{
	    for (i = 0; i < nframes; i++)
	      for (chn = 0; chn < channels; chn++)
		buffer[pos++]=in[chn][i];
	  }
	  if(jack_ringbuffer_write_space(rb)<sizeof(sample_t*)){
	    overruns_serious++;
	  }else{
	    jack_ringbuffer_write(rb,(void*)&buffer,sizeof(sample_t*));
	  }
	}


	// signal disk thread
	if (pthread_mutex_trylock (&disk_thread_lock) == 0) {
	    pthread_cond_signal (&data_ready);
	    pthread_mutex_unlock (&disk_thread_lock);
	}

	return 0;
}







/////////////////////////////////////////////////////////////////////
/////////////////// JACK CONNECTIONS ////////////////////////////////
/////////////////////////////////////////////////////////////////////

static volatile int dontreorderyet=1;


static int findnumconnections(const char **connections){
  int ret=0;
  while(connections && connections[ret]!=NULL)
    ret++;
  return ret;
}

static int compaire(const void *a, const void *b){
  return strcmp((const char*)a,(const char*)b);
}

void connect_ports_in_thread(jack_client_t *client){

  // First check if anything needs to be done, return if not.
  //printf("connect_ports\n");
  {
    int ch=0;
    const char **outportnames;

    outportnames=jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);

    ch=0;
    while(outportnames[ch]!=NULL && ch<channels){
      int lokke = 0;
      //fprintf(stderr,"ai %x\n",jack_port_by_name(client,outportnames[ch]));
      const char **connections1 = jack_port_get_all_connections(client,jack_port_by_name(client,outportnames[ch]));
      const char **connections2 = jack_port_get_all_connections(client,ports[ch]);
      int memb1=findnumconnections(connections1);
      int memb2=findnumconnections(connections2);

      if(memb1!=memb2)
	goto somethingneedstobedone;
      //      printf("same size\n");

      qsort(connections1,memb1,sizeof(char*),compaire);
      qsort(connections2,memb1,sizeof(char*),compaire);

      for(lokke=0;lokke<memb1;lokke++){
	//printf("connect_ports \"%s\" \"%s\" \n",connections1[lokke],connections2[lokke]);
	if(strcmp(connections1[lokke],connections2[lokke]))
	  goto somethingneedstobedone;
      }

      ch++;
    }
  }
  return;


 somethingneedstobedone:

  printf("(Reconnecting ports)\n");

  // disconnect
  //outportnames=jack_get_ports(client,"jack_capture*",NULL,0);
  {
    int ch;
    for(ch=0;ch<channels;ch++){
      int lokke = 0;
      //fprintf(stderr,"ch: %d %s %x %x\n",ch,outportnames[ch],ports[ch],jack_port_by_name(client,outportnames[ch]));
      const char **connections = jack_port_get_all_connections(client,ports[ch]);
      //fprintf(stderr,"ch2: %d %x\n",ch,connections);
      while(connections && connections[lokke] != NULL){
	//printf("hm %s %s\n",connections[lokke],jack_port_name(ports[ch]));
	jack_disconnect(client,connections[lokke],jack_port_name(ports[ch]));
	lokke++;
      }
    }
  }

  // connect
  {
    const char **outportnames=jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
    int ch=0;
    while(outportnames[ch]!=NULL && ch<channels){
      int lokke = 0,err;
      //fprintf(stderr,"ai %x\n",jack_port_by_name(client,outportnames[ch]));
      const char **connections = jack_port_get_all_connections(client,jack_port_by_name(client,outportnames[ch]));
      //fprintf(stderr,"%s %x..\n",outportnames[ch],connections);
      while(connections && connections[lokke] != NULL){
	//fprintf(stderr,"%s<<<\n",connections[lokke]);
	if ((err=jack_connect (client, connections[lokke], jack_port_name (ports[ch])))) {
	  fprintf (stderr, "cannot connect input port %s to %s, errorcode %s\n", jack_port_name (ports[ch]), connections[lokke],strerror(err));
	  //jack_client_close (client);
	  //exit (1);
	} 
	
	lokke++;
      }
      ch++;
    }
  }

}


void* connection_thread(void *arg){
  jack_client_t *client;

  pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  pthread_mutex_lock (&connect_thread_lock);

  if ((client = jack_client_new ("jack_connect_capture")) == 0) {
    fprintf (stderr, "jack server not running?\n");
    exit (1);
  }

  connect_ports_in_thread(client);
  dontreorderyet=0;

  can_process = 1;		/* process() can start, now */

  while(1){
    pthread_cond_wait (&connect_ready, &connect_thread_lock);
    if (!is_running) {
      goto done;
    }
    connect_ports_in_thread(client);
  }

 done:
  jack_client_close (client);
  pthread_mutex_unlock (&connect_thread_lock);
  printf ("connection thread finished\n");
  return NULL;
}


void wake_up_connection_thread(void){
  // Don't want trylock here.
  pthread_cond_signal (&connect_ready);
}



int graphordercallback(void *arg){
  if( ! (dontreorderyet==1))
    wake_up_connection_thread();
  return 0;
}




void setup_ports(void){
	unsigned int i;

	/* Allocate data structures that depend on the number of ports. */
	ports = (jack_port_t **) calloc (sizeof (jack_port_t *),channels);
	rb = jack_ringbuffer_create(JC_MIN(512*128,200+num_buffers+rb_size/jack_buffer_size));

	/* When JACK is running realtime, jack_activate() will have
	 * called mlockall() to lock our pages into memory.  But, we
	 * still need to touch any newly allocated pages before
	 * process() starts using them.  Otherwise, a page fault could
	 * create a delay that would force JACK to shut us down. */
	memset(rb->buf, 0, rb->size);

	for (i = 0; i < channels; i++) {
		char name[500];

		sprintf (name, "input%d", i+1);

		if ((ports[i] = jack_port_register (client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0)) == 0) {
			fprintf (stderr, "cannot register input port \"%s\"!\n", name);
			jack_client_close (client);
			exit (1);
		}
	}

	pthread_create (&connect_thread, NULL, connection_thread, NULL);       
}






/////////////////////////////////////////////////////////////////////
/////////////////// INIT / SHUTDOWN /////////////////////////////////
/////////////////////////////////////////////////////////////////////

void do_exit(int close_jack){
  is_running=0;

  stop_disk_thread();

  if(close_jack)
    jack_client_close(client);

  jack_ringbuffer_free (rb);

  wake_up_connection_thread();  //Tell connect thread to quit as well.
  pthread_join(connect_thread, NULL); //Wait for it.
}



void jack_shutdown (void *arg){
  fprintf (stderr, "JACK shutdown.\n");
  do_exit(0);
  exit (0);
  //abort();
}



int main (int argc, char *argv[]){

  // Arguments
  {
    OPTARGS_BEGIN("Usage: jack_capture  [--bitdepth n] [--bufsize n] [--channels n] [--leading-zeros n] [filename]\n"
		  "                     [ -b]          [ -B]         [ -c]          [ -z n]\n");
    OPTARG("--bitdepth","-b") printf("gakk\n");bitdepth = OPTARG_GETINT();
    OPTARG("--bufsize","-B") rb_size = OPTARG_GETINT();
    OPTARG("--channels","-c") channels = OPTARG_GETINT();
    OPTARG("--leading-zeros","-z") leading_zeros = JC_MIN(8,OPTARG_GETINT());
    OPTARG_LAST() filename=OPTARG_GETSTRING();
    OPTARGS_END;
  }


  // Filename
  {
    if(filename==NULL){
      int try=0;
      filename=calloc(1,5000);
      for(;;){
	const char *temp[20]={"jack_capture_%d.wav","jack_capture_%01d.wav","jack_capture_%02d.wav","jack_capture_%03d.wav","jack_capture_%04d.wav","jack_capture_%05d.wav","jack_capture_%06d.wav","jack_capture_%07d.wav","jack_capture_%08d.wav","jack_capture_%09d.wav"};
	
	sprintf(filename,temp[leading_zeros+1],++try);
	if(access(filename,F_OK)) break;
      }
      
    }
  }
  

  // Start jack
  {
    if ((client = jack_client_new ("jack_capture")) == 0) {
      fprintf (stderr, "jack server not running?\n");
      exit (1);
    }
    jack_buffer_size=jack_get_buffer_size(client);
  }


  //
  {
    buffers_init();
  }


  //
  {
    setup_disk_thread ();
  }


  // Init jack
  {
    jack_set_process_callback (client, process, NULL);
    jack_on_shutdown (client, jack_shutdown, NULL);
    
    
    jack_set_graph_order_callback(client,graphordercallback,NULL);
    
    if (jack_activate (client)) {
      fprintf (stderr, "cannot activate client");
    }
    
    setup_ports ();
  }


  //
  {
    start_disk_thread ();
  }


  // Run, wait for keyboard
  {
    printf("Recording to \"%s\". Press <Return> to stop it.\n",filename);
    {
      char gakk[64];
      fgets(gakk,49,stdin);
    }
    printf("Please wait while writing all data to disk. (shouldn't take long)\n");
    
  }


  //
  {
    do_exit(1);
  }

  
  return 0;
}


